{-# OPTIONS -#include "Curses_hsc.h" #-}
{-# LINE 1 "Curses.hsc" #-}
module Curses (
{-# LINE 2 "Curses.hsc" #-}
    --------------------------------------------------------------------
    
    Window,      -- data Window deriving Eq
    stdScr,      -- :: Window
    initScr,     -- :: IO Window
    cBreak,      -- :: Bool -> IO ()
    raw,         -- :: Bool -> IO ()
    echo,        -- :: Bool -> IO ()
    nl,          -- :: Bool -> IO ()
    intrFlush,   -- :: Bool -> IO ()
    keypad,      -- :: Window -> Bool -> IO ()
    initCurses,  -- :: IO ()
    endWin,      -- :: IO ()
    resizeTerminal,
    
    --------------------------------------------------------------------
    
    scrSize,          -- :: IO (Int, Int)
    refresh,          -- :: IO ()
    
    --------------------------------------------------------------------
    
    hasColors,      -- :: IO Bool
    startColor,     -- :: IO ()
    Pair(..),       -- newtype Pair = Pair Int deriving (Eq, Ord, Ix)
    colorPairs,     -- :: IO Int
    Color(..),      -- newtype Color = Color Int deriving (Eq, Ord, Ix)
    colors,         -- :: IO Int
    black, red, green, yellow, blue, magenta, cyan, white, -- :: Color
    initPair,       -- :: Pair -> Color -> Color -> IO ()
    pairContent,    -- :: Pair -> IO (Color, Color)
    canChangeColor, -- :: IO Bool
    initColor,      -- :: Color -> (Int, Int, Int) -> IO ()
    colorContent,   -- :: Color -> IO (Int, Int, Int)
    
    --------------------------------------------------------------------
    
    Attr,  -- data Attr deriving Eq
    attr0, -- :: Attr
    
    isAltCharset, isBlink, isBold, isDim, isHorizontal, isInvis,
    isLeft, isLow, isProtect, isReverse, isRight, isStandout, isTop,
    isUnderline, isVertical,
        -- :: Attr -> Bool
    
    setAltCharset, setBlink, setBold, setDim, setHorizontal, setInvis,
    setLeft, setLow, setProtect, setReverse, setRight, setStandout,
    setTop, setUnderline, setVertical,
        -- :: Attr -> Bool -> Attr
    
    attrSet, -- :: Attr -> Pair -> IO ()
    attrOn, attrOff,
    
    --------------------------------------------------------------------

    wAddStr, 
    addLn,         -- :: IO ()
    mvWAddStr,
    wMove,
    
    --------------------------------------------------------------------
    
    bkgrndSet,      -- :: Attr -> Pair -> IO ()
    erase,          -- :: IO ()
    wclear,         -- :: Window -> IO ()
    clrToEol,       -- :: IO ()
    move,           -- :: Int -> Int -> IO ()

    -- Cursor Routines
    CursorVisibility(..), 
    withCursor,

    standout,standend,
    attrDim, attrBold,
    attrDimOn, attrDimOff,
    attrBoldOn, attrBoldOff,
    wAttrOn,
    wAttrOff, 
    touchWin,
    --------------------------------------------------------------------
    -- Mouse Routines
    withMouseEventMask,
    ButtonEvent(..),
    MouseEvent(..),
    
    --------------------------------------------------------------------
    
    Key(..),
    getCh, 
    newPad, pRefresh, delWin, newWin,
    wClrToEol,
    withProgram,
    beep, wAttrSet, wAttrGet,

    cursesSigWinch,
    cursesTest
    ) 
    
    --------------------------------------------------------------------
    where

import Prelude hiding (pi)
import Monad 
import Char           (chr, ord, isPrint)
import Ix             (Ix)

import Data.Bits
import Control.Concurrent
import Foreign
import CForeign

import System.IO.Unsafe
import Control.Exception hiding(block)

import GenUtil(foldl')
import System.Posix.Signals
import CWString



{-# LINE 122 "Curses.hsc" #-}


------------------------------------------------------------------------

fi = fromIntegral

throwIfErr :: Num a => String -> IO a -> IO a
--throwIfErr name act = do
--    res <- act
--    if res == (#const ERR)
--        then ioError (userError ("Curses: "++name++" failed"))
--        else return res
throwIfErr s = throwIf (== (-1)) (\a -> "Curses[" ++ show a ++ "]:"  ++ s)
{-# LINE 135 "Curses.hsc" #-}

throwIfErr_ :: Num a => String -> IO a -> IO ()
throwIfErr_ name act = void $ throwIfErr name act

------------------------------------------------------------------------

data WindowTag 
type Window = Ptr WindowTag

stdScr :: Window
stdScr = unsafePerformIO (peek stdscr)
foreign import ccall "static my_curses.h &stdscr" stdscr :: Ptr Window


initScr :: IO Window
initScr = throwIfNull "initscr" initscr
foreign import ccall unsafe "my_curses.h initscr" initscr :: IO Window

cBreak :: Bool -> IO ()
cBreak True  = throwIfErr_ "cbreak"   cbreak
cBreak False = throwIfErr_ "nocbreak" nocbreak
foreign import ccall unsafe "my_curses.h cbreak" cbreak :: IO CInt
foreign import ccall unsafe "my_curses.h nocbreak" nocbreak :: IO CInt

raw :: Bool -> IO ()
raw False = throwIfErr_ "noraw" noraw
raw True  = throwIfErr_ "raw"   raw_c
foreign import ccall unsafe "my_curses.h noraw" noraw :: IO CInt
foreign import ccall unsafe "my_curses.h raw" raw_c :: IO CInt

echo :: Bool -> IO ()
echo False = throwIfErr_ "noecho" noecho
echo True  = throwIfErr_ "echo"   echo_c
foreign import ccall unsafe "my_curses.h noecho" noecho :: IO CInt
foreign import ccall unsafe "my_curses.h echo" echo_c :: IO CInt

nl :: Bool -> IO ()
nl True  = throwIfErr_ "nl"   nl_c
nl False = throwIfErr_ "nonl" nonl
foreign import ccall unsafe "my_curses.h nl" nl_c :: IO CInt
foreign import ccall unsafe "my_curses.h nonl" nonl :: IO CInt

intrFlush :: Bool -> IO ()
intrFlush bf =
    throwIfErr_ "intrflush" $ intrflush stdScr (if bf then 1 else 0)
foreign import ccall unsafe "my_curses.h intrflush" intrflush :: Window -> (Word8) -> IO CInt
{-# LINE 181 "Curses.hsc" #-}

keypad :: Window -> Bool -> IO ()
keypad win bf =
    throwIfErr_ "keypad" $ keypad_c win (if bf then 1 else 0)
foreign import ccall unsafe "my_curses.h keypad" keypad_c :: Window -> (Word8) -> IO CInt
{-# LINE 186 "Curses.hsc" #-}

foreign import ccall unsafe "my_curses.h leaveok" leaveok_c :: Window -> (Word8) -> IO CInt
{-# LINE 188 "Curses.hsc" #-}

leaveOk True = leaveok_c stdScr 1
leaveOk False = leaveok_c stdScr 0



initCurses :: IO ()
initCurses = do
    initScr
    b <- hasColors
    when b startColor
    cBreak True
    echo False
    nl False
    leaveOk True
    intrFlush False
    keypad stdScr True

endWin :: IO ()
endWin = throwIfErr_ "endwin" endwin
foreign import ccall unsafe "my_curses.h endwin" endwin :: IO CInt

------------------------------------------------------------------------

scrSize :: IO (Int, Int)
scrSize = do
    lines <- peek linesPtr
    cols  <- peek colsPtr
    return (fromIntegral lines, fromIntegral cols)

foreign import ccall "my_curses.h &LINES" linesPtr :: Ptr CInt
foreign import ccall "my_curses.h &COLS" colsPtr :: Ptr CInt


refresh :: IO ()
refresh = throwIfErr_ "refresh" refresh_c
foreign import ccall unsafe "my_curses.h refresh" refresh_c :: IO CInt

------------------------------------------------------------------------

hasColors :: IO Bool
hasColors = liftM (/= 0) has_colors
foreign import ccall unsafe "my_curses.h has_colors" has_colors :: IO (Word8)
{-# LINE 231 "Curses.hsc" #-}

startColor :: IO ()
startColor = throwIfErr_ "start_color" start_color
foreign import ccall unsafe start_color :: IO CInt

newtype Pair = Pair Int deriving (Eq, Ord, Ix)

colorPairs :: IO Int
colorPairs = fmap fromIntegral $ peek colorPairsPtr


foreign import ccall "my_curses.h &COLOR_PAIRS" colorPairsPtr :: Ptr CInt

newtype Color = Color Int deriving (Eq, Ord, Ix)

colors :: IO Int
colors = liftM fromIntegral $ peek colorsPtr

foreign import ccall "my_curses.h &COLORS" colorsPtr :: Ptr CInt

black, red, green, yellow, blue, magenta, cyan, white :: Color
black   = Color (0)
{-# LINE 253 "Curses.hsc" #-}
red     = Color (1)
{-# LINE 254 "Curses.hsc" #-}
green   = Color (2)
{-# LINE 255 "Curses.hsc" #-}
yellow  = Color (3)
{-# LINE 256 "Curses.hsc" #-}
blue    = Color (4)
{-# LINE 257 "Curses.hsc" #-}
magenta = Color (5)
{-# LINE 258 "Curses.hsc" #-}
cyan    = Color (6)
{-# LINE 259 "Curses.hsc" #-}
white   = Color (7)
{-# LINE 260 "Curses.hsc" #-}

initPair :: Pair -> Color -> Color -> IO ()
initPair (Pair p) (Color f) (Color b) =
    throwIfErr_ "init_pair" $
        init_pair (fromIntegral p) (fromIntegral f) (fromIntegral b)
foreign import ccall unsafe init_pair :: CShort -> CShort -> CShort -> IO CInt

pairContent :: Pair -> IO (Color, Color)
pairContent (Pair p) =
    alloca $ \fPtr ->
    alloca $ \bPtr -> do
        throwIfErr "pair_content" $ pair_content (fromIntegral p) fPtr bPtr
        f <- peek fPtr
        b <- peek bPtr
        return (Color (fromIntegral f), Color (fromIntegral b))
foreign import ccall unsafe pair_content :: CShort -> Ptr CShort -> Ptr CShort -> IO CInt

canChangeColor :: IO Bool
canChangeColor = liftM (/= 0) can_change_color
foreign import ccall unsafe can_change_color :: IO (Word8)
{-# LINE 280 "Curses.hsc" #-}

initColor :: Color -> (Int, Int, Int) -> IO ()
initColor (Color c) (r, g, b) =
    throwIfErr_ "init_color" $
        init_color (fromIntegral c) (fromIntegral r) (fromIntegral g) (fromIntegral b)
foreign import ccall unsafe init_color :: CShort -> CShort -> CShort -> CShort -> IO CInt

colorContent :: Color -> IO (Int, Int, Int)
colorContent (Color c) =
    alloca $ \rPtr ->
    alloca $ \gPtr ->
    alloca $ \bPtr -> do
        throwIfErr "color_content" $ color_content (fromIntegral c) rPtr gPtr bPtr
        r <- peek rPtr
        g <- peek gPtr
        b <- peek bPtr
        return (fromIntegral r, fromIntegral g, fromIntegral b)
foreign import ccall unsafe color_content :: CShort -> Ptr CShort -> Ptr CShort -> Ptr CShort -> IO CInt

foreign import ccall unsafe "my_curses.h hs_curses_color_pair" colorPair :: Pair -> (Word32)
{-# LINE 300 "Curses.hsc" #-}

{-# LINE 301 "Curses.hsc" #-}

-------------
-- Attributes 
-------------

foreign import ccall unsafe "my_curses.h attr_set" attr_set :: Attr -> CShort -> Ptr a -> IO Int
-- foreign import ccall unsafe "my_curses.h attr_get" :: Attr -> CShort -> Ptr a -> IO Int

foreign import ccall unsafe "my_curses.h wattr_set" wattr_set :: Window -> Attr -> CInt -> Ptr a -> IO CInt
foreign import ccall unsafe "my_curses.h wattr_get" wattr_get :: Window -> Ptr Attr -> Ptr CShort -> Ptr a -> IO CInt

foreign import ccall "my_curses.h attr_on" attr_on :: (Word32) -> Ptr a -> IO Int
{-# LINE 313 "Curses.hsc" #-}
foreign import ccall "my_curses.h attr_off" attr_off :: (Word32) -> Ptr a -> IO Int
{-# LINE 314 "Curses.hsc" #-}
foreign import ccall "my_curses.h attron" attron :: Int -> IO Int
foreign import ccall "my_curses.h attroff" attroff :: Int -> IO Int
foreign import ccall unsafe "my_curses.h wattron" wattron :: Window -> CInt -> IO CInt
foreign import ccall unsafe "my_curses.h wattroff" wattroff :: Window -> CInt -> IO CInt
foreign import ccall standout :: IO Int
foreign import ccall standend :: IO Int

wAttrSet :: Window -> (Attr,Pair) -> IO ()
wAttrSet w (a,(Pair p)) = throwIfErr_ "wattr_set" $ wattr_set w a (fromIntegral p) nullPtr
wAttrGet :: Window -> IO (Attr,Pair)
wAttrGet w =  alloca $ \pa -> alloca $ \pp -> (throwIfErr_ "wattr_get" $ wattr_get w pa pp nullPtr) >> (peek pa >>=  \a -> peek pp >>=  \p -> return (a,Pair (fromIntegral p)))


newtype Attr = Attr (Word32) deriving (Eq,Storable,Bits, Num, Show)
{-# LINE 328 "Curses.hsc" #-}

attr0 :: Attr
attr0 = Attr (0)
{-# LINE 331 "Curses.hsc" #-}

isAltCharset, isBlink, isBold, isDim, isHorizontal, isInvis, isLeft,
    isLow, isProtect, isReverse, isRight, isStandout, isTop,
    isUnderline, isVertical
    :: Attr -> Bool
isAltCharset = isAttr (4194304)
{-# LINE 337 "Curses.hsc" #-}
isBlink      = isAttr (524288)
{-# LINE 338 "Curses.hsc" #-}
isBold       = isAttr (2097152)
{-# LINE 339 "Curses.hsc" #-}
isDim        = isAttr (1048576)
{-# LINE 340 "Curses.hsc" #-}
isHorizontal = isAttr (33554432)
{-# LINE 341 "Curses.hsc" #-}
isInvis      = isAttr (8388608)
{-# LINE 342 "Curses.hsc" #-}
isLeft       = isAttr (67108864)
{-# LINE 343 "Curses.hsc" #-}
isLow        = isAttr (134217728)
{-# LINE 344 "Curses.hsc" #-}
isProtect    = isAttr (16777216)
{-# LINE 345 "Curses.hsc" #-}
isReverse    = isAttr (262144)
{-# LINE 346 "Curses.hsc" #-}
isRight      = isAttr (268435456)
{-# LINE 347 "Curses.hsc" #-}
isStandout   = isAttr (65536)
{-# LINE 348 "Curses.hsc" #-}
isTop        = isAttr (536870912)
{-# LINE 349 "Curses.hsc" #-}
isUnderline  = isAttr (131072)
{-# LINE 350 "Curses.hsc" #-}
isVertical   = isAttr (1073741824)
{-# LINE 351 "Curses.hsc" #-}

isAttr :: (Word32) -> Attr -> Bool
{-# LINE 353 "Curses.hsc" #-}
isAttr bit (Attr a) = a .&. bit /= 0

setAltCharset, setBlink, setBold, setDim, setHorizontal, setInvis,
    setLeft, setLow, setProtect, setReverse, setRight, setStandout,
    setTop, setUnderline, setVertical
    :: Attr -> Bool -> Attr
setAltCharset = setAttr (4194304)
{-# LINE 360 "Curses.hsc" #-}
setBlink      = setAttr (524288)
{-# LINE 361 "Curses.hsc" #-}
setBold       = setAttr (2097152)
{-# LINE 362 "Curses.hsc" #-}
setDim        = setAttr (1048576)
{-# LINE 363 "Curses.hsc" #-}
setHorizontal = setAttr (33554432)
{-# LINE 364 "Curses.hsc" #-}
setInvis      = setAttr (8388608)
{-# LINE 365 "Curses.hsc" #-}
setLeft       = setAttr (67108864)
{-# LINE 366 "Curses.hsc" #-}
setLow        = setAttr (134217728)
{-# LINE 367 "Curses.hsc" #-}
setProtect    = setAttr (16777216)
{-# LINE 368 "Curses.hsc" #-}
setReverse    = setAttr (262144)
{-# LINE 369 "Curses.hsc" #-}
setRight      = setAttr (268435456)
{-# LINE 370 "Curses.hsc" #-}
setStandout   = setAttr (65536)
{-# LINE 371 "Curses.hsc" #-}
setTop        = setAttr (536870912)
{-# LINE 372 "Curses.hsc" #-}
setUnderline  = setAttr (131072)
{-# LINE 373 "Curses.hsc" #-}
setVertical   = setAttr (1073741824)
{-# LINE 374 "Curses.hsc" #-}

setAttr :: (Word32) -> Attr -> Bool -> Attr
{-# LINE 376 "Curses.hsc" #-}
setAttr bit (Attr a) False = Attr (a .&. complement bit)
setAttr bit (Attr a) True  = Attr (a .|.            bit)

attrSet :: Attr -> Pair -> IO ()
attrSet attr (Pair p) = throwIfErr_ "attrset" $
    attr_set attr (fromIntegral p) nullPtr

attrOn :: Attr -> IO ()
attrOn (Attr attr) = throwIfErr_ "attr_on" $
    attr_on attr nullPtr


attrOff :: Attr -> IO ()
attrOff (Attr attr) = throwIfErr_ "attr_off" $
    attr_off attr nullPtr



wAttrOn :: Window -> Int -> IO ()
wAttrOn w x = throwIfErr_ "wattron" $ wattron w (fi x)

wAttrOff :: Window -> Int -> IO ()
wAttrOff w x = throwIfErr_ "wattroff" $ wattroff w (fi x)

attrDimOn :: IO ()
attrDimOn  = throwIfErr_ "attron A_DIM" $
    attron (1048576) 
{-# LINE 403 "Curses.hsc" #-}

attrDimOff :: IO ()
attrDimOff = throwIfErr_ "attroff A_DIM" $
    attroff (1048576) 
{-# LINE 407 "Curses.hsc" #-}

attrBoldOn :: IO ()
attrBoldOn  = throwIfErr_ "attron A_BOLD" $
    attron (2097152) 
{-# LINE 411 "Curses.hsc" #-}

attrBoldOff :: IO ()
attrBoldOff = throwIfErr_ "attroff A_BOLD" $
    attroff (2097152) 
{-# LINE 415 "Curses.hsc" #-}


attrDim :: Int
attrDim = (1048576)
{-# LINE 419 "Curses.hsc" #-}
attrBold :: Int
attrBold = (2097152)
{-# LINE 421 "Curses.hsc" #-}

------------------------------------------------------------------------


mvWAddStr :: Window -> Int -> Int -> String -> IO ()
mvWAddStr w y x str = wMove w y x >> wAddStr w str 

addLn :: IO ()
addLn = wAddStr stdScr "\n" 

sanifyOutput :: String -> String
sanifyOutput = map f . filter (/= '\r') where
    f c | isPrint c  = c
    f c = '~'


{-# LINE 444 "Curses.hsc" #-}

wAddStr :: Window -> String -> IO ()
wAddStr w str =  withLCStringLen (sanifyOutput str) (\(ws,len) -> throwIfErr_ ("waddnstr: " ++ show len ++ " " ++ show str) $ waddnstr w ws (fi len))
    
foreign import ccall unsafe waddnstr :: Window -> CString -> CInt -> IO CInt


{-# LINE 451 "Curses.hsc" #-}

------------------------------------------------------------------------


{-# LINE 457 "Curses.hsc" #-}

bkgrndSet :: Attr -> Pair -> IO ()
bkgrndSet (Attr a) p = bkgdset $
    fromIntegral (ord ' ') .|.
    (if a .&. 4194304 /= 0 then 4194304 else 0) .|.
{-# LINE 462 "Curses.hsc" #-}
    (if a .&. 524288 /= 0 then 524288 else 0) .|.
{-# LINE 463 "Curses.hsc" #-}
    (if a .&. 2097152 /= 0 then 2097152 else 0) .|.
{-# LINE 464 "Curses.hsc" #-}
    (if a .&. 1048576 /= 0 then 1048576 else 0) .|.
{-# LINE 465 "Curses.hsc" #-}
    (if a .&. 8388608 /= 0 then 8388608 else 0) .|.
{-# LINE 466 "Curses.hsc" #-}
    (if a .&. 16777216 /= 0 then 16777216 else 0) .|.
{-# LINE 467 "Curses.hsc" #-}
    (if a .&. 262144 /= 0 then 262144 else 0) .|.
{-# LINE 468 "Curses.hsc" #-}
    (if a .&. 65536 /= 0 then 65536 else 0) .|.
{-# LINE 469 "Curses.hsc" #-}
    (if a .&. 131072 /= 0 then 131072 else 0) .|.
{-# LINE 470 "Curses.hsc" #-}
    colorPair p
foreign import ccall unsafe bkgdset :: (Word32) -> IO ()
{-# LINE 472 "Curses.hsc" #-}

erase :: IO ()
erase = throwIfErr_ "erase" $ werase_c  stdScr
foreign import ccall unsafe "werase" werase_c :: Window -> IO CInt

wclear :: Window -> IO ()
wclear w = throwIfErr_ "wclear" $ wclear_c  w
foreign import ccall unsafe "wclear" wclear_c :: Window -> IO CInt



clrToEol :: IO ()
clrToEol = throwIfErr_ "clrtoeol" clrtoeol
foreign import ccall unsafe clrtoeol :: IO CInt

move :: Int -> Int -> IO ()
move y x =
    throwIfErr_ "move" $ move_c (fromIntegral y) (fromIntegral x)
foreign import ccall unsafe "move" move_c :: CInt -> CInt -> IO CInt

wMove :: Window -> Int -> Int -> IO ()
wMove w y x = throwIfErr_ "wmove" $ wmove w (fi y) (fi x)
foreign import ccall unsafe  wmove :: Window -> CInt -> CInt -> IO CInt

------------------
-- Cursor routines
------------------

data CursorVisibility = CursorInvisible | CursorVisible | CursorVeryVisible

vis_c :: CursorVisibility -> CInt
vis_c vis = case vis of
    CursorInvisible   -> 0
    CursorVisible     -> 1
    CursorVeryVisible -> 2
    
foreign import ccall unsafe "my_curses.h curs_set" curs_set :: CInt -> IO CInt

cursSet 0 = leaveOk True >> curs_set 0
cursSet n = leaveOk False >> curs_set n 

withCursor :: CursorVisibility -> IO a -> IO a
withCursor nv action = Control.Exception.bracket (cursSet (vis_c nv)) (\v -> case v of 
		(-1) -> return 0
{-# LINE 516 "Curses.hsc" #-}
		x -> cursSet x) (\_ -> action)



------------------------------------------------------------------------


touchWin :: Window -> IO ()
touchWin w = throwIfErr_ "touchwin" $ touchwin w
foreign import ccall touchwin :: Window -> IO CInt

newPad :: Int -> Int -> IO Window
newPad nlines ncols = throwIfNull "newpad" $ newpad (fromIntegral nlines) (fromIntegral ncols)

pRefresh :: Window -> Int -> Int -> Int -> Int -> Int -> Int -> IO ()
pRefresh pad pminrow pmincol sminrow smincol smaxrow smaxcol = throwIfErr_ "prefresh" $
    prefresh pad (fromIntegral pminrow) (fromIntegral pmincol) (fromIntegral sminrow) (fromIntegral smincol) (fromIntegral smaxrow) (fromIntegral smaxcol)

delWin :: Window -> IO ()
delWin w = throwIfErr_ "delwin" $ delwin w
    
foreign import ccall unsafe prefresh :: Window -> CInt -> CInt -> CInt -> CInt -> CInt -> CInt -> IO CInt
foreign import ccall unsafe newpad :: CInt -> CInt -> IO Window
foreign import ccall unsafe delwin :: Window -> IO CInt


newWin :: Int -> Int -> Int -> Int -> IO Window
newWin nlines ncolumn begin_y begin_x = throwIfNull "newwin" $ newwin (fi nlines) (fi ncolumn) (fi begin_y) (fi begin_x)

foreign import ccall unsafe newwin :: CInt -> CInt -> CInt -> CInt -> IO Window


wClrToEol :: Window -> IO ()
wClrToEol w = throwIfErr_ "wclrtoeol" $ wclrtoeol w
foreign import ccall unsafe wclrtoeol :: Window -> IO CInt




foreign import ccall threadsafe getch :: IO CInt


foreign import ccall unsafe def_prog_mode :: IO CInt
foreign import ccall unsafe reset_prog_mode :: IO CInt


withProgram :: IO a -> IO a
withProgram action = withCursor CursorVisible $ Control.Exception.bracket_ (def_prog_mode >> endWin) reset_prog_mode action


foreign import ccall unsafe "my_curses.h beep" c_beep :: IO CInt
foreign import ccall unsafe "my_curses.h flash" c_flash :: IO CInt

beep :: IO ()
beep = do
    br <- c_beep
    when (br /= (0)) (c_flash >> return ()) 
{-# LINE 573 "Curses.hsc" #-}


---------------
-- Key Routines
---------------

data Key
    = KeyChar Char | KeyBreak | KeyDown | KeyUp | KeyLeft | KeyRight
    | KeyHome | KeyBackspace | KeyF Int | KeyDL | KeyIL | KeyDC
    | KeyIC | KeyEIC | KeyClear | KeyEOS | KeyEOL | KeySF | KeySR
    | KeyNPage | KeyPPage | KeySTab | KeyCTab | KeyCATab | KeyEnter
    | KeySReset | KeyReset | KeyPrint | KeyLL | KeyA1 | KeyA3
    | KeyB2 | KeyC1 | KeyC3 | KeyBTab | KeyBeg | KeyCancel | KeyClose
    | KeyCommand | KeyCopy | KeyCreate | KeyEnd | KeyExit | KeyFind
    | KeyHelp | KeyMark | KeyMessage | KeyMove | KeyNext | KeyOpen
    | KeyOptions | KeyPrevious | KeyRedo | KeyReference | KeyRefresh
    | KeyReplace | KeyRestart | KeyResume | KeySave | KeySBeg
    | KeySCancel | KeySCommand | KeySCopy | KeySCreate | KeySDC
    | KeySDL | KeySelect | KeySEnd | KeySEOL | KeySExit | KeySFind
    | KeySHelp | KeySHome | KeySIC | KeySLeft | KeySMessage | KeySMove
    | KeySNext | KeySOptions | KeySPrevious | KeySPrint | KeySRedo
    | KeySReplace | KeySRight | KeySRsume | KeySSave | KeySSuspend
    | KeySUndo | KeySuspend | KeyUndo | KeyResize | KeyMouse | KeyUnknown Int
    deriving (Eq,Show,Ord)

decodeKey :: CInt -> Key
decodeKey key = case key of
    _ | key >= 0 && key <= 255 -> KeyChar (chr (fromIntegral key))
    (257)         -> KeyBreak
{-# LINE 602 "Curses.hsc" #-}
    (258)          -> KeyDown
{-# LINE 603 "Curses.hsc" #-}
    (259)            -> KeyUp
{-# LINE 604 "Curses.hsc" #-}
    (260)          -> KeyLeft
{-# LINE 605 "Curses.hsc" #-}
    (261)         -> KeyRight
{-# LINE 606 "Curses.hsc" #-}
    (262)          -> KeyHome
{-# LINE 607 "Curses.hsc" #-}
    (263)     -> KeyBackspace
{-# LINE 608 "Curses.hsc" #-}
    _ | key >= (264) && key <= (327)
{-# LINE 609 "Curses.hsc" #-}
                               -> KeyF (fromIntegral (key - 264))
{-# LINE 610 "Curses.hsc" #-}
    (328)            -> KeyDL
{-# LINE 611 "Curses.hsc" #-}
    (329)            -> KeyIL
{-# LINE 612 "Curses.hsc" #-}
    (330)            -> KeyDC
{-# LINE 613 "Curses.hsc" #-}
    (331)            -> KeyIC
{-# LINE 614 "Curses.hsc" #-}
    (332)           -> KeyEIC
{-# LINE 615 "Curses.hsc" #-}
    (333)         -> KeyClear
{-# LINE 616 "Curses.hsc" #-}
    (334)           -> KeyEOS
{-# LINE 617 "Curses.hsc" #-}
    (335)           -> KeyEOL
{-# LINE 618 "Curses.hsc" #-}
    (336)            -> KeySF
{-# LINE 619 "Curses.hsc" #-}
    (337)            -> KeySR
{-# LINE 620 "Curses.hsc" #-}
    (338)         -> KeyNPage
{-# LINE 621 "Curses.hsc" #-}
    (339)         -> KeyPPage
{-# LINE 622 "Curses.hsc" #-}
    (340)          -> KeySTab
{-# LINE 623 "Curses.hsc" #-}
    (341)          -> KeyCTab
{-# LINE 624 "Curses.hsc" #-}
    (342)         -> KeyCATab
{-# LINE 625 "Curses.hsc" #-}
    (343)         -> KeyEnter
{-# LINE 626 "Curses.hsc" #-}
    (344)        -> KeySReset
{-# LINE 627 "Curses.hsc" #-}
    (345)         -> KeyReset
{-# LINE 628 "Curses.hsc" #-}
    (346)         -> KeyPrint
{-# LINE 629 "Curses.hsc" #-}
    (347)            -> KeyLL
{-# LINE 630 "Curses.hsc" #-}
    (348)            -> KeyA1
{-# LINE 631 "Curses.hsc" #-}
    (349)            -> KeyA3
{-# LINE 632 "Curses.hsc" #-}
    (350)            -> KeyB2
{-# LINE 633 "Curses.hsc" #-}
    (351)            -> KeyC1
{-# LINE 634 "Curses.hsc" #-}
    (352)            -> KeyC3
{-# LINE 635 "Curses.hsc" #-}
    (353)          -> KeyBTab
{-# LINE 636 "Curses.hsc" #-}
    (354)           -> KeyBeg
{-# LINE 637 "Curses.hsc" #-}
    (355)        -> KeyCancel
{-# LINE 638 "Curses.hsc" #-}
    (356)         -> KeyClose
{-# LINE 639 "Curses.hsc" #-}
    (357)       -> KeyCommand
{-# LINE 640 "Curses.hsc" #-}
    (358)          -> KeyCopy
{-# LINE 641 "Curses.hsc" #-}
    (359)        -> KeyCreate
{-# LINE 642 "Curses.hsc" #-}
    (360)           -> KeyEnd
{-# LINE 643 "Curses.hsc" #-}
    (361)          -> KeyExit
{-# LINE 644 "Curses.hsc" #-}
    (362)          -> KeyFind
{-# LINE 645 "Curses.hsc" #-}
    (363)          -> KeyHelp
{-# LINE 646 "Curses.hsc" #-}
    (364)          -> KeyMark
{-# LINE 647 "Curses.hsc" #-}
    (365)       -> KeyMessage
{-# LINE 648 "Curses.hsc" #-}
    (366)          -> KeyMove
{-# LINE 649 "Curses.hsc" #-}
    (367)          -> KeyNext
{-# LINE 650 "Curses.hsc" #-}
    (368)          -> KeyOpen
{-# LINE 651 "Curses.hsc" #-}
    (369)       -> KeyOptions
{-# LINE 652 "Curses.hsc" #-}
    (370)      -> KeyPrevious
{-# LINE 653 "Curses.hsc" #-}
    (371)          -> KeyRedo
{-# LINE 654 "Curses.hsc" #-}
    (372)     -> KeyReference
{-# LINE 655 "Curses.hsc" #-}
    (373)       -> KeyRefresh
{-# LINE 656 "Curses.hsc" #-}
    (374)       -> KeyReplace
{-# LINE 657 "Curses.hsc" #-}
    (375)       -> KeyRestart
{-# LINE 658 "Curses.hsc" #-}
    (376)        -> KeyResume
{-# LINE 659 "Curses.hsc" #-}
    (377)          -> KeySave
{-# LINE 660 "Curses.hsc" #-}
    (378)          -> KeySBeg
{-# LINE 661 "Curses.hsc" #-}
    (379)       -> KeySCancel
{-# LINE 662 "Curses.hsc" #-}
    (380)      -> KeySCommand
{-# LINE 663 "Curses.hsc" #-}
    (381)         -> KeySCopy
{-# LINE 664 "Curses.hsc" #-}
    (382)       -> KeySCreate
{-# LINE 665 "Curses.hsc" #-}
    (383)           -> KeySDC
{-# LINE 666 "Curses.hsc" #-}
    (384)           -> KeySDL
{-# LINE 667 "Curses.hsc" #-}
    (385)        -> KeySelect
{-# LINE 668 "Curses.hsc" #-}
    (386)          -> KeySEnd
{-# LINE 669 "Curses.hsc" #-}
    (387)          -> KeySEOL
{-# LINE 670 "Curses.hsc" #-}
    (388)         -> KeySExit
{-# LINE 671 "Curses.hsc" #-}
    (389)         -> KeySFind
{-# LINE 672 "Curses.hsc" #-}
    (390)         -> KeySHelp
{-# LINE 673 "Curses.hsc" #-}
    (391)         -> KeySHome
{-# LINE 674 "Curses.hsc" #-}
    (392)           -> KeySIC
{-# LINE 675 "Curses.hsc" #-}
    (393)         -> KeySLeft
{-# LINE 676 "Curses.hsc" #-}
    (394)      -> KeySMessage
{-# LINE 677 "Curses.hsc" #-}
    (395)         -> KeySMove
{-# LINE 678 "Curses.hsc" #-}
    (396)         -> KeySNext
{-# LINE 679 "Curses.hsc" #-}
    (397)      -> KeySOptions
{-# LINE 680 "Curses.hsc" #-}
    (398)     -> KeySPrevious
{-# LINE 681 "Curses.hsc" #-}
    (399)        -> KeySPrint
{-# LINE 682 "Curses.hsc" #-}
    (400)         -> KeySRedo
{-# LINE 683 "Curses.hsc" #-}
    (401)      -> KeySReplace
{-# LINE 684 "Curses.hsc" #-}
    (402)        -> KeySRight
{-# LINE 685 "Curses.hsc" #-}
    (403)        -> KeySRsume
{-# LINE 686 "Curses.hsc" #-}
    (404)         -> KeySSave
{-# LINE 687 "Curses.hsc" #-}
    (405)      -> KeySSuspend
{-# LINE 688 "Curses.hsc" #-}
    (406)         -> KeySUndo
{-# LINE 689 "Curses.hsc" #-}
    (407)       -> KeySuspend
{-# LINE 690 "Curses.hsc" #-}
    (408)          -> KeyUndo
{-# LINE 691 "Curses.hsc" #-}

{-# LINE 692 "Curses.hsc" #-}
    (410)        -> KeyResize
{-# LINE 693 "Curses.hsc" #-}

{-# LINE 694 "Curses.hsc" #-}

{-# LINE 695 "Curses.hsc" #-}
    (409)        -> KeyMouse
{-# LINE 696 "Curses.hsc" #-}

{-# LINE 697 "Curses.hsc" #-}
    _                          -> KeyUnknown (fromIntegral key)





getCh :: IO Key
getCh = threadWaitRead 0 >> (liftM decodeKey $ throwIfErr "getch" getch)

--getCh :: IO Key
--getCh = liftM decodeKey $ throwIfErr "getch" getch

-- getCh :: IO Key
-- getCh = do
--     nodelay stdScr 1
--     --halfdelay 1
--     v <- getch
--     case v of
-- 	(#const ERR) -> yield >> getCh 
-- 	x -> return $ decodeKey x


resizeTerminal :: Int -> Int -> IO ()


{-# LINE 722 "Curses.hsc" #-}

resizeTerminal a b = throwIfErr_ "resizeterm"  $ resizeterm (fi a) (fi b)

foreign import ccall unsafe "my_curses.h resizeterm" resizeterm :: CInt -> CInt -> IO CInt


{-# LINE 732 "Curses.hsc" #-}


cursesSigWinch :: Maybe Signal


{-# LINE 737 "Curses.hsc" #-}

cursesSigWinch = Just (28)
{-# LINE 739 "Curses.hsc" #-}


{-# LINE 745 "Curses.hsc" #-}



------------
-- Test case
------------

cursesTest :: IO ()
cursesTest = do
    initScr
    hc <- hasColors 
    when hc startColor
    ccc <- canChangeColor
    (ys,xs) <- scrSize
    cp <- colorPairs
    cs <- colors
    endWin
    putStrLn $ "ScreenSize: " ++ show (xs,ys) 
    putStrLn $ "hasColors: " ++ show hc
    putStrLn $ "canChangeColor: " ++ show ccc
    putStrLn $ "colorPairs: " ++ show cp
    putStrLn $ "colors: " ++ show cs

    


-----------------
-- Mouse Routines
-----------------

data MouseEvent = MouseEvent {
    mouseEventId :: Int, 
    mouseEventX :: Int, 
    mouseEventY :: Int, 
    mouseEventZ :: Int, 
    mouseEventButton :: [ButtonEvent]
   } deriving(Show)

data ButtonEvent = ButtonPressed Int | ButtonReleased Int | ButtonClicked Int | 
    ButtonDoubleClicked Int | ButtonTripleClicked Int | ButtonShift | ButtonControl | ButtonAlt 
	deriving(Eq,Show)

withMouseEventMask :: [ButtonEvent] -> IO a -> IO a


{-# LINE 790 "Curses.hsc" #-}

foreign import ccall unsafe "my_curses.h mousemask" mousemask :: (Word32) -> Ptr (Word32) -> IO (Word32)
{-# LINE 792 "Curses.hsc" #-}

withMouseEventMask bes action = do
    ov <- alloca (\a ->  mousemask (besToMouseMask bes) a >> peek a) 
    r <- action 
    mousemask ov nullPtr 
    return r

besToMouseMask :: [ButtonEvent] -> (Word32)
{-# LINE 800 "Curses.hsc" #-}
besToMouseMask bes = foldl' (.|.) 0 (map cb bes) where
    cb (ButtonPressed 1) = (2)
{-# LINE 802 "Curses.hsc" #-}
    cb (ButtonPressed 2) = (128)
{-# LINE 803 "Curses.hsc" #-}
    cb (ButtonPressed 3) = (8192)
{-# LINE 804 "Curses.hsc" #-}
    cb (ButtonPressed 4) = (524288)
{-# LINE 805 "Curses.hsc" #-}
    cb (ButtonReleased 1) = (1)
{-# LINE 806 "Curses.hsc" #-}
    cb (ButtonReleased 2) = (64)
{-# LINE 807 "Curses.hsc" #-}
    cb (ButtonReleased 3) = (4096)
{-# LINE 808 "Curses.hsc" #-}
    cb (ButtonReleased 4) = (262144)
{-# LINE 809 "Curses.hsc" #-}
    cb (ButtonClicked 1) = (4)
{-# LINE 810 "Curses.hsc" #-}
    cb (ButtonClicked 2) = (256)
{-# LINE 811 "Curses.hsc" #-}
    cb (ButtonClicked 3) = (16384)
{-# LINE 812 "Curses.hsc" #-}
    cb (ButtonClicked 4) = (1048576)
{-# LINE 813 "Curses.hsc" #-}
    cb ButtonShift = (33554432)
{-# LINE 814 "Curses.hsc" #-}
    cb ButtonAlt = (67108864)
{-# LINE 815 "Curses.hsc" #-}
    cb ButtonControl = (16777216)
{-# LINE 816 "Curses.hsc" #-}
    cb _ = 0



{-# LINE 823 "Curses.hsc" #-}


-------------------------
-- code graveyard
-------------------------








{-# LINE 1026 "Curses.hsc" #-}
